#include "SNTP.hpp"

// ESP32 specific
#include "esp_log.h"
#include "esp_sntp.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"

// #include "driver/ledc.h"

// C++ system level
#include <ctime>
#include <functional>
#include <cinttypes>

// project specific
// #include "Metrics.hpp"


/*extern "C" {
  void setenv(char* env, char* value, int overwrite);
  void tzset();
}*/


static const char* TAG = "SNTP";


static void ntpTask(void* evHdl) {

  std::function<void()>* syncEventHandler = (std::function<void()>*) evHdl;


  time_t now = 0;
  struct tm timeinfo = { 0 };

  do {
    vTaskDelay(100 / portTICK_PERIOD_MS);
    time(&now);
    localtime_r(&now, &timeinfo);
  } while (timeinfo.tm_year < 2000);


  (*syncEventHandler)();


  ESP_LOGI(TAG, "Time sync complete: ");

  vTaskDelete(NULL);
}



namespace SpiderLib {

  SNTP::SNTP() {

    sntp_setoperatingmode(SNTP_OPMODE_POLL);
    sntp_setservername(0, "pool.ntp.org");  // TODO: make configurable!
    sntp_init();

    setenv("TZ", "CET-1CEST,M3.5.0,M10.5.0/3", 1);
    tzset();

    xTaskCreate(ntpTask, TAG, 8192, (void*)&syncEventHandler, 0, NULL);
  }

  void SNTP::registerTimeSyncHandler(std::function<void()> f) {
    if (inSync()) {
      f();
    } else {
      std::function<void()> _f = syncEventHandler;
      syncEventHandler = [=]() {
        _f();
        f();
      };
    }
  }

  bool SNTP::inSync() {
    return false;
  }




}
